# -*- coding: utf-8 -*-
"""
Created on Fri May  1 11:52:26 2020

@author: Reuben
"""

systems = None

def set_systems(sys):
    global systems
    systems = sys

units = None

def set_units(u):
    global units
    units = u


class Quantity():
    def __init__(self, value, spec, vector):
        self.value = value
        self.spec = spec
        self.vector = vector
        
    def set_unit(self, unit):
        if unit.vector != self.vector:
            raise ValueError('Incompatible quantity type')
        self.spec = unit.spec
                
    def _in_units(self, unit=None):
        if unit is not None:
            if unit.vector != self.vector:
                raise ValueError('Incompatible quantity type')
            spec = unit.spec
            value = self.value * unit.value
        else:
            spec = self.spec
            value = self.value
        value = systems.unitise_typed(value, spec)
        return value, spec

    def in_units(self, unit=None):
        return self._to_str(*self._in_units(unit))

    def _unitise(self):
        return systems.unitise(self.value, self.spec)

    def unitise(self):
        return self._to_str(*self._unitise())

    def _base_unitise(self):
        return systems.base_unitise(self.value, self.vector)
    
    def base_unitise(self):
        return self._to_str(*self._base_unitise())
    
    def _to_str(self, value, spec):
        return ('{:0.6g}'.format(value) + ' ' 
                + units.str_spec(spec))
    
    def __str__(self):
        return self.unitise()
    
    def __repr__(self):
        return self.__str__()

    def __mul__(self, other):
        if isinstance(other, Quantity):
            return self._mul(other)
        return Quantity(self.value * other, spec=self.spec,
                        vector=self.vector)

    def __pow__(self, other, modulo=None):
        if not isinstance(other, int):
            raise ValueError('Can only raise units to integer powers.')
        return self._pow(other)
    
    def __truediv__(self, other):
        if isinstance(other, Quantity):
            return self._div(other)
        return Quantity(self.value / other, spec=self.spec,
                        vector=self.vector)

    def __rmul__(self, other):
        if isinstance(other, Quantity):
            return self._mul(other)
        return Quantity(self.value * other, spec=self.spec,
                        vector=self.vector)


    def __rtruediv__(self, other):
        if isinstance(other, Quantity):
            return self._div(other)
        return Quantity(other / self.value,
                        spec=[-u for u in self.spec],
                        vector=-self.vector)
    
    def _mul(self, other):
        spec = self.spec.copy()
        spec.extend(other.spec)
        vector = self.vector + other.vector
        return Quantity(self.value * other.value, spec=spec,
                        vector=vector)

    def _div(self, other):
        spec = self.spec.copy()
        spec.extend([-u for u in other.spec])
        vector = self.vector - other.vector
        return Quantity(self.value / other.value, spec=spec,
                        vector=vector)


    def _pow(self, other):
        new = self
        for n in range(other - 1):
            new = new._mul(self)
        return new
            


# -*- coding: utf-8 -*-
"""
Created on Thu Apr 30 18:15:23 2020

@author: Reuben
"""

import os
import ruamel.yaml as yaml
import numpy as np

from .unit import Unit

root = os.path.dirname(os.path.abspath(__file__))



class Units():
    def __init__(self, fname=None):
        self.load(fname)
    
    def _ind(self, s):
        if s in self._ind_dct:
            return self._ind_dct[s]
        index = len(self._num_dct) + 1
        self._num_dct[index] = s
        self._ind_dct[s] = index
        self._num_dct[-index] = '-' + s
        self._ind_dct['-' + s] = -index
        return index
    
    def str(self, ind):
        return self._num_dct[ind]
        
    def _add_utype(self, i, utype):
        self._add_utype[i] = utype
    
    def new(self, abbr, value, vector, spec, name, utype):
        if abbr in self.units:
            raise KeyError(abbr + ' is already defined.')
        spec = [self._ind(u) for u in spec]
        index = self._ind(abbr)
        index_base = self._ind(utype)
        self._utypes[index] = index_base
        u = Unit(abbr, value, vector, spec, name)
        self.safe_set(self.units, abbr, u)
        # Now make the corresponding inverse ('negative') unit
        ut = [-u for u in spec]
        uneg = Unit(abbr, 1/value, -vector, ut, name)
        self.safe_set(self.units, '-' + abbr, uneg)
        self._utypes[-index] = -index_base
        return u
    
    def _make_utypes(self, types):
        self.utypes = types
        def vec(ind):
            a = np.zeros(len(types))
            a[ind] = 1
            return a
        for i, t in enumerate(types):
            self.new(t, 1.0, vec(i), [t], t, t)
    
    def load(self, fname=None):
        self.units = {} # The unit instances
        self.bases = {} # The base units for time, length, etc
        self._utypes = {} # the length, time for given id
        self._num_dct = {} # The attr for given index
        self._ind_dct = {} # the index for given attr
        raw = self._load_raw(fname)
        self._make_type_dct(raw)
    
    def _load_raw(self, fname=None):
        if fname is None:
            fname = os.path.join(root, 'units') + '.yaml'
        with open(fname, 'r') as f:
            raw = yaml.safe_load(f)
        return raw
    
    def safe_set(self, unit_dct, key, val):
        if key in unit_dct:
            raise KeyError(key + ' already defined.')
        else:
            unit_dct[key] = val
        
    def _derive(self, spec):
        us = [self[u] for u in spec]
        vector = np.sum([u.vector for u in us], axis=0)
        value = np.prod([u.value for u in us])
        return value, vector
        
    def _make_unit(self, units, utype, abbr, v):
        value, derivation, name = v
        if not isinstance(derivation, list):
            derivation = [derivation]
        m, vector = self._derive(derivation)
        self.new(abbr, value * m, vector, [abbr], name, utype)
    
    def _make_type_dct(self, dct):
        units = self.units
        for utype, d in dct.items():
            if isinstance(d, list):
                self._make_utypes(d)
                continue
            for abbr, v in d.items():
                if abbr == '_base':
                    base_abbr = v
                    self.bases[self._ind(utype)] = self._ind(base_abbr)
                else:
                    self._make_unit(units, utype, abbr, v)

    def __getitem__(self, abbr):
        if abbr in self.units:
            return self.units[abbr]
        raise KeyError(str(abbr) + ' not defined')

    def __getattr__(self, abbr):
        if abbr not in ['units', 'bases'] and abbr in self.units:
            return self.units[abbr]
        else:
            return self.__getattribute__(abbr)
    
    def get_by_index(self, i):
        return self.units[self._num_dct[i]]
    
    def str_spec(self, spec):
        if spec is None:
            return 'base'
        elif len(spec)==0:
            return 'dimensionless'
        num = [i for i in spec if i > 0]
        den = [-i for i in spec if i < 0]
        def f(v, c):
            if c == 1:
                return v
            else:
                return v + str(c)
        def process(lst):
            out = [units._num_dct[i] for i in lst]    
            out.sort()
            d = {v: out.count(v) for v in out}
            return [f(v, c) for v, c in d.items()]
        num = process(num)
        den = process(den)
        s_num = '1' if len(num) == 0 else '.'.join(num)
        n = len(den)
        if n == 0:
            return s_num
        elif n == 1:
            return s_num + '/' + den[0]
        else:
            return s_num + '/(' + '.'.join(den) + ')'

units = Units()


        
    


# -*- coding: utf-8 -*-
"""
Created on Fri May  1 11:52:42 2020

@author: Reuben
"""

from . import base
from . import system
from . import quantity
from . import unit

def init():
    global initialised, base
    quantity.set_systems(system.systems)
    quantity.set_units(base.units)
    initialised = True
    
init()

units = base.units
set_system = system.set_system


# -*- coding: utf-8 -*-
"""
Created on Fri May  1 11:20:07 2020

@author: Reuben
"""

import os
import ruamel.yaml as yaml

from . import base

root = os.path.dirname(os.path.abspath(__file__))


class Systems():
    def __init__(self, fname=None):
        self.load(fname)
    
    def load(self, fname=None):
        raw = self._load_raw(fname)
        self._sys_dct = self._make_sys_dct(raw)
        for name in self._sys_dct.keys():
            self._active = name
            break

    def _load_raw(self, fname=None):
        if fname is None:
            fname = os.path.join(root, 'systems') + '.yaml'
        with open(fname, 'r') as f:
            raw = yaml.safe_load(f)
        return raw
    
    def _make_sys_dct(self, raw):
        return {n: System(dct) for n, dct in raw.items()}
    
    def set_system(self, name):
        self._active = name
    
    @property
    def active(self):
        return self._sys_dct[self._active]
        
    def unitise(self, val, utype):
        return self._sys_dct[self._active].unitise(val, utype)

    def base_unitise(self, val, type_vec):
        return self._sys_dct[self._active].base_unitise(val, type_vec)

    def unitise_typed(self, val, spec):
        return self._sys_dct[self._active].unitise_typed(val, spec)


class System():
    def __init__(self, dct):
        self._sys_dct = self._make_sys_dct(dct)
    
    def _make_sys_dct(self, dct):
        d = {}
        for spec, units_raw in dct.items():
            unit_dct = {}
            units_raw.reverse()
            for abbr in units_raw:
                if isinstance(abbr, list):
                    unit = base.units[abbr[1]]
                    mult = abbr[0] * unit.value
                    a = base.units._ind(unit.abbr)
                else:
                    unit = base.units[abbr]
                    mult = unit.value
                    a = base.units._ind(abbr)
                unit_dct[a] = mult
            d[base.units._ind(spec)] = unit_dct
        return d
    
    def calc_utypes(self, vector):
        utypes = []
        for n, name in zip(vector, base.units.utypes):
            i = base.units._ind(name)
            if n > 0:
                utypes.extend([i]*int(abs(n)))
            else:
                utypes.extend([-i]*int(abs(n)))
        return utypes

    def _unitise_one(self, val, spec):
        if abs(spec) not in self._sys_dct:
            return val, spec
        d = self._sys_dct[abs(spec)]
        div = spec < 0
        trials = []
        i_vals = []
        for i, mult in d.items():
            if div:
                trials.append(val * mult)
                i_vals.append(-i)
            else:
                trials.append(val / mult)
                i_vals.append(i)
        a = [max(abs(v), 1/abs(v)) for v in trials]
        ind = a.index(min(a))
        return trials[ind], i_vals[ind]

    def _base_unitise_one(self, val, spec):
        div = False
        if spec < 0:
            div = True
        utype_i = abs(base.units._utypes[spec]) # length, force, etc
        if utype_i in base.units.bases:
            b = base.units.bases[utype_i] # m, N etc
        else:
            b = utype_i
        u = base.units.get_by_index(b)
        if div:
            return val * u.value, -b
        return val / u.value, b
    
    def unitise(self, val, spec):
        new_val = val
        out_spec = []
        utypes = base.units._utypes
        utype = [utypes[s] for s in spec]
        for u in utype:
            new_val, ut = self._unitise_one(new_val, u)
            out_spec.append(ut)
        return new_val, out_spec
    
    def base_unitise(self, val, vector):
        base_spec = self.calc_utypes(vector)
        new_val = val
        spec = []
        for u in base_spec:
            new_val, ut = self._base_unitise_one(new_val, u)
            spec.append(ut)
        return new_val, spec
    
    def unitise_typed(self, val, spec):
        out = val
        for u in spec:
            out /= base.units.get_by_index(u).value
        return out
    

systems = Systems()        
active = systems.active
set_system = systems.set_system




# -*- coding: utf-8 -*-
"""
Created on Fri May  1 11:51:58 2020

@author: Reuben
"""

from .quantity import Quantity

class Unit(Quantity):
    def __init__(self, abbr, value, vector, spec, name):
        self.abbr = abbr
        self.value = value
        self.vector = vector
        self.spec = spec
        self.name = name
        
    def __str__(self):
        return self.base_unitise()

    def __repr__(self):
        return self.__str__()

    def __rlshift__(self, other):
        if not isinstance(other, Quantity):
            raise ValueError('Quantity not recognised.')
        other.set_unit(self)

    def __rrshift__(self, other):
        return other / self.value
    
    def __rmatmul__(self, other):
        return Quantity(self.value * other, spec=self.spec,
                        vector=self.vector)
    